/*
 * Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <inttypes.h>
#include <string.h>

#include <ucg_delay.h>
#include <ucg_irq.h>
#include <ucg_callout.h>
#include <ucg_reent_intr.h>

static volatile uint32_t global_ms;
static volatile uint8_t brightness; /* 0 to 100 */
static volatile int8_t brightness_inc = 1;
static struct ucg_callout_mgr *p_intr_cm = NULL;

#if defined(UCGINE_ARCH_stm32)

#if defined(UCGINE_SUBARCH_stm32f3)
#include <stm32f30x.h>
#include <stm32f3_discovery.h>
#elif defined(UCGINE_SUBARCH_stm32f4)
#include <stm32f4xx.h>
#include <stm32f4_discovery.h>
#endif

static void target_init(void)
{
	RCC_ClocksTypeDef RCC_Clocks;

	/* Get SYSCLK, HCLK and PCLKx frequency */
	RCC_GetClocksFreq(&RCC_Clocks);
	/* generate an interrupt every ms */
	SysTick_Config(RCC_Clocks.HCLK_Frequency / 1000);

	/* Initialize LEDs */
	STM_EVAL_LEDInit(LED6);
}

static void led_on(void)
{
	STM_EVAL_LEDOn(LED6);
}

static void led_off(void)
{
	STM_EVAL_LEDOff(LED6);
}

void reent_intr(void)
{
	global_ms++;
	if (p_intr_cm != NULL)
		ucg_callout_manage(p_intr_cm);
}

/* called every ms */
__attribute__ ((naked)) void SysTick_Handler(void)
{
	UCG_REENT_INTR(reent_intr);
}
#elif defined(UCGINE_ARCH_avr)
#include <avr/io.h>

static void led_on(void)
{
	PORTB |= (1 << 5);
}

static void led_off(void)
{
	PORTB &= (~(1 << 5));
}

SIGNAL(TIMER0_OVF_vect)
{
	static uint16_t i = 0;

	i++;
	if ((i & 0x3) != 0)
		return;

	global_ms++;
	if (p_intr_cm != NULL)
		ucg_callout_manage(p_intr_cm);
}

static void target_init(void)
{
	DDRB = (1 << 5);

	TCCR0B = (1 << CS01);   /* div = 8 */
	TIMSK0 |= (1 << TOIE0); /* enable timer0 intr */
}
#endif

static uint16_t get_time_ms(void)
{
	ucg_irqflags_t flags;
	uint32_t ms;

	flags = ucg_irq_lock_save();
	ms = global_ms;
	ucg_irq_unlock_restore(flags);
	return ms;
}

/* every ms, enable or disable the led depending on brightness */
static void led_control_cb(struct ucg_callout_mgr *cm,
	struct ucg_callout *tim, void *arg)
{
	static uint8_t accu;

	(void)arg;

	accu += brightness;
	if (accu < 100) {
		led_off();
	} else {
		led_on();
		accu -= 100;
	}

	ucg_callout_reschedule(cm, tim, 1);
}

/* every 50 ms, change increase or decrease the brightness from 0 to 100 */
static void led_update_brightness_cb(struct ucg_callout_mgr *cm,
	struct ucg_callout *tim, void *arg)
{
	(void)arg;

	brightness += brightness_inc;
	if (brightness == 0)
		brightness_inc = 1;
	else if (brightness == 100)
		brightness_inc = -1;

	ucg_callout_reschedule(cm, tim, 10);
}

/* every second, sleep during 100ms */
static void sleep_cb(struct ucg_callout_mgr *cm,
	struct ucg_callout *tim, void *arg)
{
	(void)arg;

	ucg_delay_ms(200);
	ucg_callout_reschedule(cm, tim, 300);
}

int main(void)
{
	struct ucg_callout_mgr intr_cm;
	struct ucg_callout_mgr loop_cm;
	struct ucg_callout br_timer;
	struct ucg_callout led_timer;
	struct ucg_callout sleep_timer;
	int i;

	target_init();

	/* toggle the pin before starting */
	for (i = 0; i < 3; i++) {
		ucg_delay_ms(500);
		ucg_delay_ms(500);
	}

	/* init loop callout manager */
	ucg_callout_mgr_init(&loop_cm, get_time_ms);

	/* load a timer in the loop cmgr that will update the brightness */
	ucg_callout_init(&br_timer, led_update_brightness_cb, NULL, 128);
	ucg_callout_schedule(&loop_cm, &br_timer, 0);

	/* init intr callout manager */
	ucg_callout_mgr_init(&intr_cm, get_time_ms);
	ucg_irq_lock();
	p_intr_cm = &intr_cm;
	ucg_irq_unlock();

	/* load a timer in the intr cmgr that control led on/off */
	ucg_callout_init(&led_timer, led_control_cb, NULL, 128);
	ucg_callout_schedule(&intr_cm, &led_timer, 0);

	/* load a lower prio timer in the intr cmgr that just sleeps */
	ucg_callout_init(&sleep_timer, sleep_cb, NULL, 100);
	ucg_callout_schedule(&intr_cm, &sleep_timer, 0);

	while (1) {
		ucg_callout_manage(&loop_cm);
	}

	while (1);
	return 0;
}
